"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = transformRewriteImports;
exports.defaultRecognizedExtensions = void 0;
require("core-js/modules/es.array.push.js");
require("core-js/modules/es.error.cause.js");
var _package = require("../../package.json");
var _debug = _interopRequireDefault(require("debug"));
var _template = _interopRequireDefault(require("@babel/template"));
var util = _interopRequireWildcard(require("@babel/types"));
function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }
function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const debugNamespace = `${_package.name}:index`;
const debug = (0, _debug.default)(debugNamespace);
const globalMetadata = {};
const defaultRecognizedExtensions = ['.js', '.jsx', '.mjs', '.cjs', '.json'];
exports.defaultRecognizedExtensions = defaultRecognizedExtensions;
let reporterTimeout;
function transformRewriteImports() {
  let rewriteOptionsObjIdentifier = undefined;
  let rewriteFnIdentifier = undefined;
  return {
    name: 'transform-rewrite-imports',
    visitor: {
      Program: {
        enter() {
          rewriteOptionsObjIdentifier = undefined;
          rewriteFnIdentifier = undefined;
          clearTimeout(reporterTimeout);
        },
        exit(_, state) {
          if (process.env.NODE_ENV != 'test') {
            reporterTimeout = setTimeout(() => {
              if (!state.opts.silent) {
                let totalGlobalImports = 0;
                let totalTransformedImports = 0;
                let totalFiles = 0;
                const globalMetadataEntries = Object.entries(globalMetadata);
                globalMetadataEntries.forEach(([path, metadata]) => {
                  if (state.opts.verbose) {
                    console.log(`Rewrote ${metadata.transformedImports.length} of ${metadata.totalImports} imports in file ${path}`);
                    metadata.transformedImports.forEach(transformedImport => {
                      console.log(`  ${transformedImport}`);
                    });
                  }
                  totalGlobalImports += metadata.totalImports;
                  totalTransformedImports += metadata.transformedImports.length;
                  totalFiles += 1;
                });
                if (!(state.opts.appendExtension || state.opts.replaceExtensions)) {
                  console.log(`(${_package.name} was loaded without configuration, making it a noop)`);
                } else {
                  console.log(`${state.opts.verbose && globalMetadataEntries.length ? '---\n' : ''}Rewrote ${totalTransformedImports} of ${totalGlobalImports} imports ${totalFiles == 1 ? 'in 1 file' : `across ${totalFiles} files`}`);
                }
              }
            }, 100);
          }
        }
      },
      ImportDeclaration: declarationHandler,
      ExportAllDeclaration: declarationHandler,
      ExportNamedDeclaration: declarationHandler,
      CallExpression(path, state) {
        var _path$node$callee, _path$node$callee2, _path$node$callee3, _path$node$arguments;
        const isDynamicImport = ((_path$node$callee = path.node.callee) === null || _path$node$callee === void 0 ? void 0 : _path$node$callee.type) == 'Import';
        const isRequire = ((_path$node$callee2 = path.node.callee) === null || _path$node$callee2 === void 0 ? void 0 : _path$node$callee2.type) == 'Identifier' && ((_path$node$callee3 = path.node.callee) === null || _path$node$callee3 === void 0 ? void 0 : _path$node$callee3.name) == 'require';
        const firstArgument = (_path$node$arguments = path.node.arguments) === null || _path$node$arguments === void 0 ? void 0 : _path$node$arguments[0];
        const firstArgumentIsStringLiteral = (firstArgument === null || firstArgument === void 0 ? void 0 : firstArgument.type) == 'StringLiteral';
        const {
          appendExtension,
          recognizedExtensions,
          replaceExtensions
        } = {
          ...state.opts,
          recognizedExtensions: state.opts.recognizedExtensions || defaultRecognizedExtensions
        };
        if (isDynamicImport || isRequire) {
          const metadata = getLocalMetadata(state);
          metadata.totalImports += 1;
          if (appendExtension || replaceExtensions) {
            const specifierType = isDynamicImport ? 'dynamic import' : 'require';
            if (!firstArgument) {
              console.warn(`[${getFilenameFromState(state)}]: warning: a ${specifierType} statement has no arguments!`);
            } else if (firstArgumentIsStringLiteral) {
              const importPath = firstArgument.value;
              const rewrittenPath = rewrite(importPath, {
                appendExtension,
                recognizedExtensions,
                replaceExtensions
              });
              if (importPath == rewrittenPath) {
                debug(`[${getFilenameFromState(state)}]: ${specifierType} evaluated but unchanged: "${importPath}"`);
              } else {
                path.node.arguments[0] = util.stringLiteral(rewrittenPath);
                const debugString = `[CallExpression]: ${specifierType} "${importPath}" => "${rewrittenPath}"`;
                metadata.transformedImports.push(debugString);
                debug(debugString);
              }
            } else {
              const globalScope = path.scope.getProgramParent();
              if (!rewriteFnIdentifier) {
                const rewriteFnAst = _template.default.expression.ast(rewrite.toString());
                rewriteFnIdentifier = globalScope.generateUidIdentifier('_rewrite');
                globalScope.push({
                  id: rewriteFnIdentifier,
                  init: rewriteFnAst,
                  kind: 'const'
                });
              }
              if (!rewriteOptionsObjIdentifier) {
                const rewriteOptionsObjAst = _template.default.expression.ast(JSON.stringify({
                  appendExtension,
                  recognizedExtensions,
                  replaceExtensions
                }));
                rewriteOptionsObjIdentifier = globalScope.generateUidIdentifier('_rewrite_options');
                globalScope.push({
                  id: rewriteOptionsObjIdentifier,
                  init: rewriteOptionsObjAst,
                  kind: 'const'
                });
              }
              path.get('arguments')[0].replaceWith(util.callExpression(rewriteFnIdentifier, [firstArgument, rewriteOptionsObjIdentifier]));
              const debugString = `[CallExpression]: ${specifierType} first argument wrapped with rewrite function`;
              metadata.transformedImports.push(debugString);
              debug(debugString);
            }
          }
        }
      }
    }
  };
}
function declarationHandler(path, state) {
  const metadata = getLocalMetadata(state);
  metadata.totalImports += 1;
  if (path.node.source && (state.opts.appendExtension || state.opts.replaceExtensions)) {
    const importPath = path.node.source.value;
    const specifierType = path.node.type.startsWith('Import') ? 'import' : 'export';
    const rewrittenPath = rewrite(importPath, {
      ...state.opts,
      recognizedExtensions: state.opts.recognizedExtensions || defaultRecognizedExtensions
    });
    if (importPath == rewrittenPath) {
      debug(`[${getFilenameFromState(state)}]: ${specifierType} evaluated but unchanged: "${importPath}"`);
    } else {
      path.node.source = util.stringLiteral(rewrittenPath);
      const debugString = `[declarationHandler]: ${specifierType} "${importPath}" => "${rewrittenPath}"`;
      metadata.transformedImports.push(debugString);
      debug(debugString);
    }
  }
}
function getLocalMetadata(state) {
  const key = getFilenameFromState(state);
  return globalMetadata[key] = globalMetadata[key] || {
    totalImports: 0,
    transformedImports: []
  };
}
function getFilenameFromState(state) {
  return state.filename || '<no path>';
}
const rewrite = (importPath, options) => {
  if (typeof importPath != 'string') {
    throw new TypeError(`rewrite error: expected import specifier of type string, not ${typeof importPath}`);
  }
  let finalImportPath = importPath;
  let replacementMap;
  replacementMap = undefined;
  if (options.replaceExtensions) {
    Object.entries(options.replaceExtensions).some(([target, replacement]) => {
      if (target.startsWith('^') || target.endsWith('$')) {
        const targetRegExp = new RegExp(target);
        if (targetRegExp.test(finalImportPath)) {
          replacementMap = [targetRegExp, replacement];
          return true;
        }
      } else if (finalImportPath.endsWith(target)) {
        replacementMap = [target, replacement];
        return true;
      }
    });
  }
  if (replacementMap) {
    const [target, replacement] = replacementMap;
    finalImportPath = finalImportPath.replace(target, replacement);
  }
  const isRelative = finalImportPath.startsWith('./') || finalImportPath.startsWith('../') || finalImportPath == '.' || finalImportPath == '..';
  if (options.appendExtension && isRelative) {
    const endsWithSlash = finalImportPath.endsWith('/');
    if (/(^\.?\.\/?$)|(\/\.?\.\/?$)/.test(finalImportPath)) {
      finalImportPath += `${endsWithSlash ? '' : '/'}index${options.appendExtension}`;
    } else {
      const hasRecognizedExtension = !endsWithSlash && options.recognizedExtensions.some(extension => {
        return finalImportPath.endsWith(extension);
      });
      if (!hasRecognizedExtension) {
        finalImportPath = endsWithSlash ? finalImportPath + `index${options.appendExtension}` : finalImportPath + options.appendExtension;
      }
    }
  }
  return finalImportPath;
};